package com.anthonycicchetti.slackbot

import com.anthonycicchetti.slackbot.osrs.ApiItem
import com.anthonycicchetti.slackbot.osrs.getOsrsItemPrice
import com.anthonycicchetti.slackbot.osrs.getOsrsItemByName
import com.anthonycicchetti.slackbot.utility.*
import io.javalin.http.Context
import io.javalin.Javalin
import io.javalin.plugin.openapi.OpenApiOptions
import io.javalin.plugin.openapi.OpenApiPlugin
import io.javalin.plugin.openapi.dsl.document
import io.javalin.plugin.openapi.dsl.documented
import io.javalin.plugin.openapi.ui.SwaggerOptions
import io.swagger.v3.oas.models.OpenAPI
import io.swagger.v3.oas.models.info.Info
import io.swagger.v3.oas.models.servers.Server
import kotlinx.coroutines.runBlocking
import org.slf4j.LoggerFactory
import java.util.*
import kotlin.math.roundToInt

fun main() {
    val logger = LoggerFactory.getLogger("main")

    val app = Javalin.create { config ->
        config.registerPlugin(OpenApiPlugin(createOpenApiOptions()))
        config.enableWebjars()
        config.enableCorsForAllOrigins()
        config.requestLogger { ctx, ms ->
            logger.info("Took ${ms.roundToInt()} to process ${ctx.req.requestURL}")
        }
    }.start(7000)
    with(app) {
        options("/*", documented(document().ignore()) {})
        get("/", documented(document().ignore()) { ctx -> ctx.result("Hello World") })

        post("/slack/receive", documented(document().operation {
            it.description = "NOT DOCUMENTED. Required for slack usage"
            it.summary = "NOT DOCUMENTED. Required for slack usage"
        }.ignore()) { ctx ->
            logger.info("Received request from ${ctx.req.requestURL}")
            handleSlackEvent(ctx)
        })

        post("/api/v1/spongebob", documented(document().operation {
            it.description = "Responds with the ${"spongemocked".toSpongemock()} version of the string posted"
            it.summary = "POST for ${"spongemocking".toSpongemock()}"
        }.body<String>().json<TextResponse>("200")) { ctx ->
            logger.info("Spongemock: Received request from ${ctx.req.requestURL}")
            ctx.json(TextResponse(200, ctx.body().toSpongemock()))
        })

        post("/api/v1/uppercase", documented(document().operation {
            it.description = "Responds with the UPPERCASED version of the string posted"
            it.summary = "POST for UPPERCASING"
        }.body<String>().json<TextResponse>("200")) { ctx ->
            logger.info("Uppercase: Received request from ${ctx.req.requestURL}")
            ctx.json(TextResponse(200, ctx.body().uppercase(Locale.getDefault())))
        })

        post("/api/v1/claptext", documented(document().operation {
            it.description = "Responds \uD83D\uDC4F with \uD83D\uDC4F the \uD83D\uDC4F claptext \uD83D\uDC4F version"
            it.summary = "POST \uD83D\uDC4F for \uD83D\uDC4F claptext"
        }.body<String>().json<TextResponse>("200")) { ctx ->
            logger.info("Claptext: Received request from ${ctx.req.requestURL}")
            ctx.json(TextResponse(200, ctx.body().toClapText()))
        })

        post("/api/v1/osrsitem", documented(document().operation {
            it.description = "Accepts an OSRS item name and searches for its prices on the wiki"
            it.summary = "POST for OSRS Price"
        }.body<String>().json<BlockResponse>("200")) { ctx ->
            runBlocking {
                logger.info("OSRS Item: Received request for ${ctx.body()}")
                val item: ApiItem? = getOsrsItemByName(ctx.body())
                if (item == null) {
                    ctx.json(notFoundItem)
                } else {
                    val itemPrices = getOsrsItemPrice(item)
                    ctx.json((BlockResponse(200, item.name, itemPrices.high, itemPrices.low)))
                }
            }
        })
    }


    Runtime.getRuntime().addShutdownHook(Thread {
        app.stop()
    })

    app.events { event ->
        event.serverStopping { logger.info("Server stopping") }
        event.serverStopped { logger.info("Server stopped") }
    }
}

private fun handleSlackEvent(ctx: Context) {
    // Short circuit for ssl check
    if ((ctx.formParam("ssl_check") ?: "0") == "1") {
        ctx.status(200)
        return
    }

    val responseObj = RespObj(
        command = ctx.formParamMap()["command"]?.get(0).processToCommand(),
        text = ctx.formParamMap()["text"]?.get(0) ?: "",
        response_url = ctx.formParamMap()["response_url"]?.get(0) ?: "",
        team_id = ctx.formParamMap()["team_id"]?.get(0) ?: "",
        channel_id = ctx.formParamMap()["channel_id"]?.get(0) ?: ""
    )

    sendResponse(ctx, responseObj)
}

private fun String?.processToCommand(): Commands {
    return when (this) {
        "/spongemock2" -> Commands.Spongebob
        "/uppercase" -> Commands.Uppercase
        "/clapback" -> Commands.Claptext
        "/osrsitem" -> Commands.OsrsItem
        else -> Commands.Error
    }
}

private fun sendResponse(ctx: Context, respObj: RespObj) {
    val returnMap = mutableMapOf<String, String>()
    with(returnMap) {
        when (respObj.command) {
            Commands.Spongebob -> {
                put("text", respObj.text.toSpongemock()); put("response_type", "in_channel")
            }

            Commands.Uppercase -> {
                put("text", respObj.text.uppercase(Locale.getDefault())); put("response_type", "in_channel")
            }

            Commands.Claptext -> {
                put("text", respObj.text.toClapText()); put("response_type", "in_channel")
            }

            Commands.OsrsItem -> runBlocking {
                val item = getOsrsItemByName(respObj.text)
                val itemPricesBlock: BlockResponse = if (item == null) {
                    notFoundItem
                } else {
                    val itemPrices = getOsrsItemPrice(item)
                    BlockResponse(200, item.name, itemPrices.high, itemPrices.low)
                }
                val logger = LoggerFactory.getLogger("sendResponse")
                logger.info(itemPricesBlock.response)
                put("blocks", itemPricesBlock.response); put("response_type", "in_channel")
            }

            Commands.Error -> {
                put("text", respObj.text); put("response_type", "ephemeral")
            }
        }

    }
    ctx.json(returnMap)
}

private fun createOpenApiOptions(): OpenApiOptions {
    val o = {
        OpenAPI()
            .info(Info().version("0.3.0").description("Slackbot Open API documentation"))
            .addServersItem(Server().url("https://acicchetti.dev/"))
    }

    return OpenApiOptions(o)
        .path("/swagger-docs")
        .swagger(SwaggerOptions("/swagger").title("Slackbot Swagger Docs"))
}